local MAJOR, MINOR = "LibMindReader", 1

if LibStub == nil then
  error(MAJOR .. ' requires LibStub');
end

local mr, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if not mr then return end -- No Upgrade needed.

mr.embeds = mr.embeds or {} -- what objects embed this lib

-- Get the helper functions
local hp = LibStub("LibHandyHelper"):Embed(mr.hp);

--- embedding and embed handling
local mixins = {
	"Build", "GetTempStorage", "AddFtionTip", "AddFtionTips"
}

-- mr:Embed( target )
-- target (object) - target object to embed LibMindReader in
--
-- Embeds LibMindReader into the target object making the functions from the mixins list available on target:..
function mr:Embed(target)
	for k, v in pairs(mixins) do
		if target[v] then
			error('Cannot mixin '..v..' member already exists');
		end
		target[v] = self[v];
	end
	self.embeds[target] = true
	return target
end

-- mr:OnEmbedDisable( target )
-- target (object) - target object that is being disabled
--
function mr:OnEmbedDisable(target)
	-- No clue what to do here, but other libs seem to have this...
	-- Perhaps I should clean up target's logs
	if temporary[target] then
		hp:del(temporary[target]);
		temporary[target] = nil;
	end
end

--[[
	LibMindReader functions
--]]

-- Table to hold the built tables in
local temporary = hp:new();

-- table to hold the function tips in
local ftips = hp:new();

--[[
:Build()
Build an AceConfigDialog3 optiontable from a variable

	var(object) = the variable
	name(string) = the name of the variable
	
returns the name under which the optiontable will be stored, the table itself and the setup-table where the config stores it's working variables
--]]
function mr:Build(var, name)

	if var == nil then
		return;
	end
	
	if temporary[self] == nil then
		temporary[self] = hp:new();
	end
	
	if name == nil then
		name = '' .. var;
	end
	-- Make a unique name
	local rootname = name .. GetTime();
	
	local t = hp:new();
		t.name = rootname;
		t.type = 'group';
		t.args = hp.new();
	
	local s = hp:new(); -- Logs all the vars already encountered, for avoiding up- and cross-links and for keeping workvars for the GUI
	local i = 100; -- Start at 100, this way the children are always AFTER the controls
	local nexti = i + 1; -- Holds the next free spot in the queue
	
	-- To avoid cross-links we work the tree level by level, always adding it's children to the end of the queue
	local q = hp:new();
	q[i] = hp:new(); -- hold the children still left to do, should hold everything needed to process
		q[i].key = name;
		q[i].val = var;
		q[i].args = t.args;
		q[i].parentval = nil;
	local c, n;
	
	while q[i] do
		c = q[i]; -- current todo
		if type(c.val) == 'table' then
			-- Check if we haven't seen this table yet
			if not s[''..c.val] then
				-- n holds the new node
				n = mr:MakeTableNode(c.name, c.parentval, c.val, s);
				n.order = i;
				c.args[c.name] = n;
				
				-- Add the current var's children to the todo
				for k, v in pairs(c.val) do
					q[nexti] = hp:new();
						q[nexti].key = k;
						q[nexti].val = v;
						q[nexti].args = n.args;
						q[nexti].parentval = c.val;
					nexti = nexti + 1;
				end
			end
		
		elseif type(c.val) == 'number' then
			n = mr:MakeNumberNode(c.name, c.parentval, c.val, s);
			n.order = i;
			c.args[c.name] = n;
		
		elseif type(c.val) == 'string' then
			n = mr:MakeStringNode(c.name, c.parentval, c.val, s);
			n.order = i;
			c.args[c.name] = n;
		
		elseif type(c.val) == 'function' then
			n = mr:MakeFunctionNode(c.name, c.parentval, c.val, s);
			n.order = i;
			c.args[c.name] = n;
		
		end
		
		-- Clean up nicely
		hp:del(q[i]);
		q[i] = nil;
		i = i + 1;
	end
	
	local nt = hp:new();
		nt.table = t;
		nt.setup = s;
	temporary[self][rootname] = nt;
	
	hp:del(q);
	
	return rootname, t, s;

end

--[[
:MakeTableNode()
Generates a single node for a table

	name(string) = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(table) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
function mr:MakeTableNode(name, parentval, val, setup)

	local sname = ''..val;
	local s = hp:new();
	setup[sname] = s;

	local n = hp:new();
		n.name = name;
		n.type = 'group';
		n.args = hp:new();
	
	-- The table has an inline group that contains controls for doing stuff on the variable
	local sg = hp:new();
		sg.name = sname;
		sg.type = 'group';
		sg.order = 1;
		sg.inline = true;
		sg.args = hp:new();
	n.args[s] = sg;
		
	local label = hp:new();
		label.name = 'Table variable';
		label.type = 'description';
		label.order = 1;
	sg.args.label = label;
	
	local execinput = hp:new();
		execinput.name = 'Code';
		execinput.type = 'input';
		execinput.order = 2;
		execinput.multiline = 8;
		execinput.arg = hp:new();
			execinput.arg.s = s;
		execinput.get = function(info)
			return info.arg.s.execinput;
		end;
		execinput.set = function(info, value)
			info.arg.s.execinput = value;
		end;
	sg.args.execinput = execinput;
		
	local execbutton = hp:new();
		execbutton.name = 'Execute as Method';
		execbutton.type = 'execute';
		execbutton.order = 3;
		execbutton.arg = hp:new();
			execbutton.arg.s = s;
			execbutton.arg.t = val;
		execbutton.func = function(info)
			local f = assert(loadstring(info.arg.s.execinput));
			f(info.arg.t);
		end;
	sg.args.execbutton = execbutton;
	
	return n;

end

--[[
:MakeNumberNode()
Generates a single node for a number

	name(string) = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(number) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
function mr:MakeNumberNode(name, parentval, val, setup)

	local sname = ''..parentval..name;

	local n = hp:new();
		n.name = name;
		n.type = 'group';
		n.args = hp:new();
	
	-- The table has an inline group that contains controls for doing stuff on the variable
	local sg = hp:new();
		sg.name = sname;
		sg.type = 'group';
		sg.order = 1;
		sg.inline = true;
		sg.args = hp:new();
	n.args[s] = sg;
		
	local label = hp:new();
		label.name = 'Number variable';
		label.type = 'description';
		label.order = 1;
	sg.args.label = label;
	
	local numberinput = hp:new();
		numberinput.name = 'Number';
		numberinput.type = 'input';
		numberinput.order = 2;
		if parentval then
			numberinput.arg = hp:new();
				numberinput.arg.p = parentval;
			numberinput.get = function(info)
				return info.arg.p[info[#info]];
			end;
			numberinput.set = function(info, value)
				info.arg.p[info[#info]] = tonumber(value);
			end;
		else
			numberinput.arg = val;
			numberinput.get = function(info)
				return info.arg;
			end;
			numberinput.disabled = true;
		end
	sg.args.numberinput = numberinput;
	
	return n;

end

--[[
:MakeStringNode()
Generates a single node for a string

	name(string) = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(string) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
function mr:MakeStringNode(name, parentval, setup)

	local sname = ''..parentval..name;

	local n = hp:new();
		n.name = name;
		n.type = 'group';
		n.args = hp:new();
	
	-- The table has an inline group that contains controls for doing stuff on the variable
	local sg = hp:new();
		sg.name = sname;
		sg.type = 'group';
		sg.order = 1;
		sg.inline = true;
		sg.args = hp:new();
	n.args[s] = sg;
		
	local label = hp:new();
		label.name = 'String variable';
		label.type = 'description';
		label.order = 1;
	sg.args.label = label;
	
	local stringinput = hp:new();
		stringinput.name = 'String';
		stringinput.type = 'input';
		stringinput.order = 2;
		stringinput.multiline = 8; -- a string can contain multiple lines
		stringinput.arg = hp:new();
			stringinput.arg.p = parentval;
		stringinput.get = function(info)
			return info.arg.p[info[#info]];
		end;
		stringinput.set = function(info, value)
			info.arg.p[info[#info]] = tonumber(value);
		end;
	sg.args.stringinput = stringinput;
	
	return n;

end

--[[
:MakeFunctionNode()
Generates a single node for a function
Uses functiontips for having more info on the paramaters to provide

	name(string) = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(function) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
function mr:MakeFunctionNode(name, parentval, val, setup)

	-- dilema: do we keep one global setup per function (as functions tend to duplicate rapidly) or one setup per link to a function
	-- I went for option no. 2
	local sname = ''..parentval..name;
	-- just in case I change my mind, take into account that the setup may already exist
	local s;
	if setup[sname] == nil then
		s = hp:new();
		setup[sname] = s;
	else
		s = setup[sname];
	end
	
	-- Build the parameter-function-creator
	local params = '';
	local paramstring = 'nil';
	if parentval then
		params = 'self';
	end
	if ftips[val] then
		paramstring = ftips[val];
	end
	s.execinput = 'return function('..params..[[)
		return ]]..paramstring..[[
	end]];

	local n = hp:new();
		n.name = name;
		n.type = 'group';
		n.args = hp:new();
	
	-- The table has an inline group that contains controls for doing stuff on the variable
	local sg = hp:new();
		sg.name = sname;
		sg.type = 'group';
		sg.order = 1;
		sg.inline = true;
		sg.args = hp:new();
	n.args[s] = sg;
		
	local label = hp:new();
		label.name = 'Table variable';
		label.type = 'description';
		label.order = 1;
	sg.args.label = label;
	
	local execinput = hp:new();
		if parentval then
			execinput.name = 'Code (first param = self)';
		else
			execinput.name = 'Code (no first param given)';
		end
		execinput.type = 'input';
		execinput.order = 2;
		execinput.multiline = 8;
		execinput.arg = hp:new();
			execinput.arg.s = s;
		execinput.get = function(info)
			return info.arg.s.execinput;
		end;
		execinput.set = function(info, value)
			info.arg.s.execinput = value;
		end;
	sg.args.execinput = execinput;
		
	local execbutton = hp:new();
		execbutton.name = 'Execute';
		execbutton.type = 'execute';
		execbutton.order = 3;
		if parentval then
			execbutton.arg = hp:new();
				execbutton.arg.s = s;
				execbutton.arg.f = val;
				execbutton.arg.p = parentval;
			execbutton.func = function(info)
				-- parse the parameter-function-creator
				local pfc = assert(loadstring(info.arg.s.execinput));
				-- Create the parameter-function
				local pf = pfc();
				-- pass the parenttable as the first argument to it to get the arguments
				-- this way the function can return it as the first parameter and call the ftion as a method
				local p = pf(info.arg.p);
				-- execute the ftion
				execbutton.arg.f(p);
			end
		else
			execbutton.arg = hp:new();
				execbutton.arg.s = s;
				execbutton.arg.f = val;
			execbutton.func = function(info)
				-- parse the parameter-function-creator
				local pfc = assert(loadstring(info.arg.s.execinput));
				-- Create the parameter-function
				local pf = pfc();
				-- No parenttable, so the function will have to provide the 'self' by itself
				local p = pf();
				-- execute the ftion
				execbutton.arg.f(p);
			end
		end
	sg.args.execbutton = execbutton;
	
	return n;

end

--[[
:GetTempStorage()

returns the table where the Build-results are stored in for the current addon
]]--
function mr:GetTempStorage()

	if temporary[self] == nil then
		temporary[self] = hp:new();
	end
	
	return temporary[self];

end

--[[
:GetAllStorage()

return the whole table where the Build-results are stored, this isn't mixed in
]]--
function mr:GetAllStorage()

	return temporary;

end

--[[
:AddFtionTip()

	func(function) = the function in question
	paramstring(string) = the parameterstring of the function
	ismethod(boolean) = indicated if there is an extra 'self' parameter to be added
	
Adds the function tip to the table
--]]
function mr:AddFtionTip(func, paramstring, ismethod)
	if ismethod then
		paramstring = 'self, '..paramstring;
	end
	ftips[func] = paramstring;
end

--[[
:AddFtionTips()

	tiptable(table) = table with a whole lotta tips
	
Adds the tips to our own tip-table
]]--
function mr:AddFtionTips(tiptable)

	for func, paramstring in pairs(tiptable) do
		ftips[func] = paramstring;
	end

end

--[[
:Show()

	rootname(string) = name of Build, may be nil, if not nil and the table is found in the storage, tab is ignored
	tab(table) = the GUI-table to show, if name is nil, this needs to have a value
	
IF AceConfigDialog3 is found this shows the provided table as a GUI. If not, it does nothing.
]]--
function mr:Show(rootname, tab)

	local acd = LibStub("AceConfigDialog-3.0");
	local ac = LibStub("AceConfig-3.0");
	if acd then
		if rootname and temporary[rootname] then
			tab = temporary[rootname];
		end
		if tab == nil then
			if name then
				error('No table found to show for '..name);
			else
				error('No table found to show');
			end
		end
		
		--ac:RegisterOptionsTable(
	end

end


--[[
	Rest of the lib-stuff
--]]

-- what's this for anyway? isn't mr.embeds empty at this time??
for addon in pairs(mr.embeds) do
	mr:Embed(addon)
end
